/*
 *             Automatically Tuned Linear Algebra Software v3.2
 *                    (C) Copyright 1999 R. Clint Whaley                     
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions, and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *   3. The name of the University of Tennessee, the ATLAS group,
 *      or the names of its contributers may not be used to endorse
 *      or promote products derived from this software without specific
 *      written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 */

#include "atlas_misc.h"
#include "atlas_level1.h"

void Mjoin(PATL,dotc_sub)(const int N, const TYPE *X, const int incX, 
                          const TYPE *Y, const int incY, TYPE *dot)
{
   int i;
   const int incx = incX<<1, incy = incY<<1;
   TYPE rdot = ATL_rzero, idot = ATL_rzero, rx, ix, ry, iy;

   if (incX == 1 && incY == 1)
   {
      for (i=N; i; i--, X += 2, Y += 2)
      {
         rx = *X;
         ix = X[1];
         ry = *Y;
         iy = Y[1];
         rdot += rx * ry + ix * iy;  
         idot += rx * iy - ix * ry;
      }
   }
   else
   {
      for (i=N; i; i--, X += incx, Y += incy)
      {
         rx = *X;
         ix = X[1];
         ry = *Y;
         iy = Y[1];
         rdot += rx * ry + ix * iy;  
         idot += rx * iy - ix * ry;
      }
   }
   *dot = rdot;
   dot[1] = idot;
}
